# -*- perl -*-

#
# $Id: Main.pm,v 1.2 2001/02/14 01:16:06 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2001 Slaven Rezic. All rights reserved.
#
# Mail: slaven.rezic@berlin.de
# WWW:  http://www.rezic.de/~eserte/
#

package ESRI::Shapefile::Main;

use 5.006;

use strict;
use vars qw(%FIELDS);

use constant FILECODE => 9994;
use constant VERSION => 1000;

BEGIN {
    for (["Null Shape" => 0],
	 ["Point"      => 1],
	 ["PolyLine"   => 3],
	 ["Polygon"    => 5],
	 ["MultiPoint" => 8],
	 ["PointZ"     => 11],
	 ["PolyLineZ"  => 13],
	 ["PolygonZ"   => 15],
	 ["MultiPointZ"=> 18],
	 ["PointM"     => 21],
	 ["PolyLineM"  => 23],
	 ["PolygonM"   => 25],
	 ["MultiPointM"=> 28],
	 ["MultiPatch" => 31],
	) {
	(my $shape_type = $_->[0]) =~ s/\W//g;
	my $cmd = 'use constant SHAPE_'.uc($shape_type).' => '.$_->[1];
	#warn $cmd;
	eval $cmd;
    }
}

use fields qw(File Root Header Records FH);

use base qw(Class::Accessor);

__PACKAGE__->mk_accessors(keys %FIELDS);

sub new {
    my ESRI::Shapefile::Main $self = shift;
    $self = fields::new($self) unless (ref $self);

    my(%args) = @_;

    if ($args{-root}) {
	$self->Root($args{-root});
    }
    if ($args{-file}) {
	$self->set_file($args{-file});
    }

    $self;
}

sub set_file {
    my($self, $file) = @_;

    $self->File($file);

    open(my $fh, $file) or die "Can't open $file: $!";
    $self->FH($fh);
    binmode $fh;

    read $fh, my($header), 100;
    $self->Header(ESRI::Shapefile::Main::Header->new($header));

    my @records;
    while(!eof $fh) {
	push @records, ESRI::Shapefile::Main::Record->new($self);
    }
    $self->Records(\@records);

    close $fh;
    $self->FH(undef);

#use Data::Dumper; print STDERR "Line " . __LINE__ . ", File: " . __FILE__ . "\n" . Data::Dumper->Dumpxs([$self],[]); # XXX

}

######################################################################

package ESRI::Shapefile::Main::Header;

use vars qw(%FIELDS);

use fields qw(FileCode FileLength Version ShapeType BoundingBox);

use base qw(Class::Accessor);

__PACKAGE__->mk_accessors(keys %FIELDS);

sub new {
    my ESRI::Shapefile::Main::Header $self = shift;
    $self = fields::new($self) unless (ref $self);

    my $buf = shift;

    $self->FileCode(unpack("N", substr($buf, 0, 4)));
    die "Wrong file code @{ $self->FileCode }"
	if $self->FileCode != ESRI::Shapefile::Main::FILECODE;
    $self->FileLength(unpack("N", substr($buf, 24, 4)));
    $self->Version(unpack("V", substr($buf, 28, 4)));
    die "Wrong version @{ $self->Version }"
	if $self->Version != ESRI::Shapefile::Main::VERSION;
    $self->ShapeType(unpack("V", substr($buf, 32, 4)));

    my %bounding_box;
    my $pos = 36;
    for (qw(Xmin Ymin Xmax Ymax Zmin Zmax Mmin Mmax)) {
	# XXX d on Bigendian?
	$bounding_box{$_} = unpack("d", substr($buf, $pos, 8));
	$pos+=8;
    }
    $self->BoundingBox(\%bounding_box);

}

######################################################################

package ESRI::Shapefile::Main::Record;

use vars qw(%FIELDS);

use fields qw(RecordNumber ContentLength ShapeType);

use base qw(Class::Accessor);

__PACKAGE__->mk_accessors(keys %FIELDS);

sub new {
    my $class = shift;
    my $root  = shift;

    read $root->FH, my($header), 12;

    my $shape_type = unpack("V", substr($header, 8, 4));
    my $new_class = "ESRI::Shapefile::Main::Record::" .
	{ ESRI::Shapefile::Main::SHAPE_POLYLINE() => "PolyLine",
	  #XXX
	}->{$shape_type};

    my $record = $new_class->new_from_file($root);

    $record->RecordNumber(unpack("N", substr($header, 0, 4)));
    # here, ContentLength is in Bytes
    $record->ContentLength(unpack("N", substr($header, 4, 4)) * 2);
    $record->ShapeType($shape_type);

    $record;
}

######################################################################

package ESRI::Shapefile::Main::Record::PolyLine;

use vars qw(%FIELDS);

use base qw(ESRI::Shapefile::Main::Record Class::Accessor);

use fields qw(Box Lines);

__PACKAGE__->mk_accessors(keys %FIELDS);

sub new_from_file {
    my ESRI::Shapefile::Main::Record::PolyLine $self = shift;
    $self = fields::new($self) unless (ref $self);

    my $root = shift;

    read $root->FH, my($header), 40;

    # XXX big endian?
    $self->Box([unpack("d4", substr($header, 0, 32))]);

    my $num_parts  = unpack("V", substr($header, 32, 4));
    my $num_points = unpack("V", substr($header, 36, 4));

    my $parts_len = $num_parts*4;
    read $root->FH, $header, $parts_len;
    my(@parts)  = unpack("V$num_parts", $header);

    read $root->FH, $header, $num_points*8*2;
    # XXX big endian?
    my(@points) = unpack("d".2*$num_points, $header);

    my @lines;
    for(my $part_i = 0; $part_i <= $#parts; $part_i++) {
	my $end_index = ($part_i < $#parts ? $parts[$part_i+1]-1 : $#points);
	my @line;
	for(my $i = $parts[$part_i]; $i < $end_index; $i+=2) {
	    push @line, [@points[$i..$i+1]];
	}
	push @lines, \@line;
    }

    $self->Lines(\@lines);

    $self;
}


1;

__END__
